import ftplib
import os
from datetime import datetime
from typing import List,Tuple


def _get_fftp_files(
    ftp: ftplib.FTP, path: str, files: list, dirs: list, file_dates: list
):

    lst: list = []

    year = datetime.now().strftime("%Y")

    ftp.dir(path, lst.append)
    for item in lst:
        parts = item.split(maxsplit=8)
        fullname = f"{path}/{parts[-1]}"
        if item.startswith("d"):
            dirs.append(fullname)
            _get_fftp_files(ftp, fullname, files, dirs, file_dates)
        else:
            if ":" in parts[7]:
                parts[7] = year
            date = datetime.strptime(f"{parts[5]} {parts[6]} {parts[7]}", "%b %d %Y")
            # datestr = date.strftime("%Y%m%d")
            files.append(fullname)
            file_dates.append(date)


def _makedirs(ftp: ftplib.FTP, path: str, remote_dirs: set):
    assert path.startswith("/")
    if path in remote_dirs:
        return

    dirs = path.split("/")[1:]

    ftp.cwd("/")
    for dir in dirs:
        try:
            ftp.cwd(dir)
            continue
        except:
            ftp.mkd(dir)
            ftp.cwd(dir)

    remote_dirs.add(path)


class FTP:
    def __init__(self, host: str, port: int, username: str, password: str):
        self._ftp = ftplib.FTP()
        self._ftp.connect(host, port)
        self._ftp.login(username, password)

    @property
    def ftp(self):
        return self._ftp

    def close(self):
        self._ftp.close()

    def list_all(self, path: str) -> Tuple[list, list, list]:
        files: list = []
        dirs: list = []
        file_dates: list = []

        _get_fftp_files(self._ftp, path, files, dirs, file_dates)
        return files, dirs, file_dates

    def push_file(self, local_path: str, remote_path: str):
        dirname = os.path.dirname(remote_path)
        self.makedirs(dirname)
        self._ftp.cwd(dirname)
        with open(local_path, "rb") as f:
            self._ftp.storbinary(f"STOR {os.path.basename(remote_path)}", f)

    def try_push_dir(self, local_path: str, remote_path: str):
        """将本地目录推送到远程目录，如果远程文件存在，则忽略"""

        assert remote_path.startswith("/")

        remote_files, remote_dirs, _ = (set(x) for x in self.list_all(remote_path))

        _makedirs(self._ftp, remote_path, remote_dirs)
        self._ftp.cwd(remote_path)

        for root, dirs, files in os.walk(local_path):
            remote_dir = os.path.relpath(root, local_path)
            remote_dir = os.path.join(remote_path, remote_dir).replace("\\", "/")
            self._ftp.cwd(remote_dir)

            # 创建目录
            for dir in dirs:
                try:
                    self._ftp.mkd(dir)
                except:
                    pass

            # 推送文件
            for file in files:
                local_full_name = os.path.join(root, file)
                local_relpath = os.path.relpath(local_full_name, local_path)
                remote_file = os.path.join(remote_path, local_relpath).replace(
                    "\\", "/"
                )
                if remote_file in remote_files:
                    continue
                with open(local_full_name, "rb") as f:
                    self._ftp.storbinary(f"STOR {file}", f)
                    print(f"ftp push:{local_relpath} -> {remote_file}")

    def makedirs(self, path: str):
        assert path.startswith("/")
        dirs = path.split("/")[1:]
        self._ftp.cwd("/")
        for dir in dirs:
            try:
                self._ftp.cwd(dir)
                continue
            except:
                self._ftp.mkd(dir)
                self._ftp.cwd(dir)



import shlex
import subprocess
import sys
from typing import Iterable, Tuple, Union


def call_binary(cmd: str) -> Tuple[bytes, bool]:
    if sys.platform == "win32":
        args = cmd
    else:
        # linux must split arguments
        args = shlex.split(cmd)
    try:
        data = subprocess.check_output(args)
        return data, True
    except subprocess.CalledProcessError as callerr:
        print(f"cmd = {cmd}, callerr.output = {callerr.output}", file=sys.stderr)
        return b"", False
    except IOError as ioerr:
        print(f"cmd = {cmd}, ioerr = {ioerr}", file=sys.stderr)
        return b"", False


# return (output, isOk)
def call(
    cmd: str, printOutput: bool = False, encoding: Union[None, str, Iterable[str]] = None
) -> Tuple[str, bool]:
    data, isOk = call_binary(cmd)
    if not isOk:
        return "", False

    if encoding is None:
        encoding = "utf-8"
    isOk = False
    output = ""
    itor = (encoding,) if isinstance(encoding, str) else encoding
    for enc in itor:
        try:
            output = data.decode(enc)
            isOk = True
            break
        except UnicodeDecodeError:
            pass
    if isOk and printOutput:
        print(output)
    return output, isOk



import csv
from typing import List, Tuple

import openpyxl


def save_csv(filename, rows: List[List], header: List = None):
    with open(filename, "w", newline="") as f:
        writer = csv.writer(f)
        if header:
            writer.writerow(header)
        for row in rows:
            writer.writerow(row)


def save_xlsx(
    filename: str, rows: List[List], header: List = None, images: List[Tuple[str, str]] = None
):
    """Save data to xlsx file. 注意图片不能设置为嵌入单元格图片。
    如需设置嵌入单元格图片，则执行以下步骤:
    1. 打开Excel
    2. 选中任意一张图片
    3. Ctrl+A 选中所有图片
    4. 右键选择嵌入单元格图片

    Args:
        filename (str): filename
        rows (list[list]):
        header (list, optional): header. Defaults to None.
        images (list[tuple[str, str]], optional): [(path, anchor),...]. Defaults to None.
    """
    wb = openpyxl.Workbook()
    ws = wb.active
    if header:
        ws.append(header)
    for row in rows:
        ws.append(row)

    if images:
        for path, anchor in images:
            img = openpyxl.drawing.image.Image(path)
            ws.add_image(img, anchor)

    wb.save(filename)



"some function for io"

import os
import shutil
from typing import List


def makedirs(filename, isfile=False):
    "like os.makedirs, can pass a filename"
    dirname = os.path.dirname(filename) if isfile else filename
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)


def write_file(filename, data, mode=None, encoding="utf-8", newline=None, append=False):
    "save data to file"
    makedirs(filename, isfile=True)
    if not mode:
        mode = "w" if isinstance(data, str) else "wb"
    if append:
        mode = mode.replace("w", "a")
    if isinstance(data, (bytearray, bytes)):
        encoding = None
        newline = None
    with open(filename, mode=mode, encoding=encoding, newline=newline) as f:
        f.write(data)


def read_file(filename, isbin=False, encoding="utf-8"):
    "read data from file"
    mode = "rb" if isbin else "r"
    if isbin:
        encoding = None
    with open(filename, mode=mode, encoding=encoding) as f:
        return f.read()


def write_lines(
    filename,
    lines: List[str],
    encoding="utf-8",
    newline="\n",
    newlineinlines=False,
    append=False,
):
    "save lines to file"
    mode = "at" if append else "wt"
    with open(filename, mode=mode, encoding=encoding) as f:
        if newlineinlines:
            f.writelines(lines)
        else:
            for line in lines:
                f.write(line)
                f.write(newline)


def read_lines(filename, encoding="utf-8", keepends=False):
    "read lines from file"
    with open(filename, mode="r", encoding=encoding) as f:
        if keepends:
            return f.readlines()
        else:
            return f.read().splitlines(keepends)


def copy_file(src, dst):
    makedirs(dst, True)
    shutil.copyfile(src, dst)


def check_encoding(filename: str, *encodings: str):
    "check encoding on file"
    for encoding in encodings:
        with open(filename, mode="r", encoding=encoding) as f:
            try:
                f.read()
                return encoding
            except UnicodeDecodeError:
                pass
    return None



"some function for misc"


from datetime import datetime
from typing import Any, Callable, Iterable, List, Union

_all = all
_any = any


def all(iterable: Iterable, func: Callable[[Any], bool] = None) -> bool:
    """like buildin all(iterable), can accept an test function"""
    if func is None:
        return _all(iterable)

    for it in iterable:
        if not func(it):
            return False

    return True


def any(iterable: Iterable, func: Callable[[Any], bool] = None) -> bool:
    """like buildin any(iterable), can accept an test function"""
    if func is None:
        return _any(iterable)

    for it in iterable:
        if func(it):
            return True

    return False


def parse_datetime(timestr: Union[str, datetime]) -> datetime:
    """
    parse str to datatime

    support format:
    - 2021-12-27 10:31:30
    - 2021/12/27 10:31:30
    - 2021/12/27 03:31
    - 2021/12/27 3:31
    """
    if isinstance(timestr, datetime):
        return timestr

    assert isinstance(
        timestr, str
    ), f"Type Error timestr must be str or datetime, timestr:{type(timestr)}"

    timearr = timestr.replace("-", " ").replace("/", " ").replace(":", " ").split(" ")
    assert len(timearr) == 5 or len(timearr) == 6
    timeintarr: List[Any] = list([int(n) for n in timearr])
    if len(timeintarr) == 5:
        timeintarr.append(0)
    return datetime(*timeintarr)



from setuptools import setup, find_packages


setup(
    name="litefeel-pycommon",
    version="0.4.15",
    description="my python common lib",
    author="litefeel",
    author_email="litefeel@gmail.com",
    url="http://github.com/litefeel/pycommon",
    license="MIT",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
    ],
    install_requires=["openpyxl"],
    packages=find_packages(exclude=["test"]),
    python_requires=">=3.8",
)
